(* absyn/absyn.sml *)

functor AbsynFn(structure MakeString : MAKESTRING
		structure Source : SOURCE
		  ) : ABSYN =
  struct

    structure Source	= Source

    (* start line, start column, end line, endcolumn *)
	datatype loc = LOC of int * int * int * int

    (* filename, pos start, pos end *)
	datatype info = INFO of string * int * int * loc ref

    datatype ident	= IDENT of string * info ref
    type var		= ident
    type tyvar		= ident

    datatype longid	= LONGID of ident option * ident * info ref

    datatype lit = CCONlit of char * info ref
				 | ICONlit of int * info ref
				 | RCONlit of real * info ref
				 | SCONlit of string * info ref

    datatype ty	= VARty of tyvar * info ref
				| CONSty of ty list * longid * info ref
				| TUPLEty of ty list * info ref
				| RELty of ty list * ty list * info ref

    datatype pat = WILDpat of info ref
				 | LITpat of lit * info ref
				 | CONpat of longid * info ref
				 | STRUCTpat of longid option * pat list * info ref
				 | BINDpat of var * pat * info ref
				 | IDENTpat of ident * pat ref * info ref

    datatype exp = LITexp of lit * info ref
				 | CONexp of longid * info ref
				 | VARexp of longid * info ref
				 | STRUCTexp of longid option * exp list * info ref
				 | IDENTexp of longid * exp ref * info ref

    datatype goal = CALLgoal of longid * exp list * pat list * info ref
				  | EQUALgoal of var * exp * info ref
				  | LETgoal of pat * exp * info ref
				  | NOTgoal of goal * info ref
				  | ANDgoal of goal * goal * info ref

    datatype result	= RETURN of exp list * info ref
					| FAIL of info ref
					
    datatype clause	= CLAUSE1 of goal option * ident * pat list * result * info ref
					| CLAUSE2 of clause * clause * info ref 

    datatype conbind = CONcb of ident * info ref
					 | CTORcb of ident * ty list * info ref

    datatype datbind = DATBIND of tyvar list * ident * conbind list * info ref

    datatype typbind = TYPBIND of tyvar list * ident * ty * info ref

    datatype relbind = RELBIND of ident * ty option * clause * info ref

    datatype spec = WITHspec of string * interface ref * info ref
				  | ABSTYPEspec of bool * tyvar list * ident * info ref
				  | TYPEspec of typbind list * info ref
				  | DATAspec of datbind list * typbind list * info ref
				  | VALspec of ident * ty * info ref
				  | RELspec of ident * ty * info ref

    and interface = INTERFACE of {modid: ident,
					specs: spec list,
					source: Source.source} * info ref

    datatype dec = WITHdec of string * interface ref * info ref
			| TYPEdec of typbind list * info ref
			| DATAdec of datbind list * typbind list * info ref
			| VALdec of ident * exp * info ref
			| RELdec of relbind list * info ref

    datatype module	= MODULE of interface * dec list * info ref

    structure IdentDict =
      AADictFn(struct
		 type ord_key = ident
		 fun compare(IDENT(s1,_), IDENT(s2,_)) =
		   String.compare(s1, s2)
	       end)

    val dummyLoc  = LOC(0, 0, 0, 0)

    val dummyInfo = INFO("RML", 0, 0, ref dummyLoc)

    fun makeIdent(name, info) = IDENT(name, info)
    fun rmlIdent name = IDENT(name, ref dummyInfo)
    fun identName(IDENT(name,_)) = name
    fun identCtxInfo(IDENT(_, ref(info))) = info
    fun lidentName(LONGID(SOME(IDENT(name1,_)),IDENT(name2,_),_)) = name1^"."^name2
    | lidentName(LONGID(NONE,IDENT(name,_),_)) = name
    fun lidentCtxInfo(LONGID(_, _, ref(info))) = info    
    fun identEqual(IDENT(s1,_), IDENT(s2,_)) = (s1 = s2)

    fun litEqual(CCONlit(c1, _), CCONlit(c2, _)) = (c1 = c2)
      | litEqual(ICONlit(i1, _), ICONlit(i2, _)) = (i1 = i2)
      | litEqual(RCONlit(r1, _), RCONlit(r2, _)) = Real.==(r1, r2)
      | litEqual(SCONlit(s1, _), SCONlit(s2, _)) = (s1 = s2)
      | litEqual(_, _) = false

    fun litString(CCONlit(c, _)) = MakeString.ccvt c
      | litString(ICONlit(i, _)) = MakeString.icvt i
      | litString(RCONlit(r, _)) = MakeString.rcvt r
      | litString(SCONlit(s, _)) = MakeString.scvt s
    
    val dummyInterface =
      INTERFACE({modid = rmlIdent "", specs = [], source = Source.dummy}, ref dummyInfo)

  end (* functor AbsynFn *)
