/**************************************************
[ rml-debug.c ] 
- creation: adrpo 2002-10
  implementation for rml debugging 
  this file contains all the functions for debugging.
  look also into runtime/debugging for:
  + debug-debug.c 
  + debug-print.c
  + debug-push-vars.c
- last modified: 2004-11-24
***************************************************/

#include <stdio.h>
#include <unistd.h>
#include <termio.h>

#include <readline/readline.h>
#include <readline/history.h>
#include <stdarg.h>

#include <signal.h>

#include <errno.h>

#include <sys/stat.h>
#include <sys/types.h>

#include <fcntl.h>

#ifndef  _RML_DEBUG_H_
#include "rml.h"
#endif

#define RML_TYPE_SUFFIX_NONE   0
#define RML_TYPE_SUFFIX_LIST   1
#define RML_TYPE_SUFFIX_OPTION 2
#define RML_TYPE_SUFFIX_LVAR   3

char *rmldb_command;
int execution_type = RMLDB_STEP;
int rmldb_show = RMLDB_HIDE;

/* adrpo 
this filter is set on the relations you want to make the 
program to stop (like a breakpoint, but with string matching  */
char relation_filter[RML_DEBUG_MAX_BREAKPOINTS][RML_DEBUG_MAX_STRING];
int number_of_filters = 0; /* of course, the number of filters. */
int depth_of_variable_print = 10; /* the depth of the variable showing */
int max_string_variable_print = 20; /* how many characters we display in strings */
int max_backtrace_entries = 100;
int rml_debug_print_livevars_each_step = RML_LIVEVARS_OFF;

/* adrpo 
this flag is set to 1 when you need a step by step execution 
if you want run execution then this flag is set on 0  */ 
int last_command = 0;

struct call_stack_node *call_stack;
struct call_stack_node *last_node;


struct rml_debug_var_node *rml_debug_var_in_start;
struct rml_debug_var_node *rml_debug_var_in_end;

struct rml_debug_var_node *rml_debug_var_out_start;
struct rml_debug_var_node *rml_debug_var_out_end;

char *rmldb_ttyname;


char rml_display_vars[RML_DEBUG_MAX_DISPLAY_VARS][RML_DEBUG_MAX_STRING];
int number_of_display_vars=0; /* of course, the number active display vars. */

rml_var_db_t *rml_var_db_start;
rml_var_db_t *rml_var_db_end;
rml_type_db_t *rml_type_db_start;
rml_type_db_t *rml_type_db_end;
rml_con_db_t *rml_con_db_start;
rml_con_db_t *rml_con_db_end;
rml_relation_db_t *rml_relation_db_start;
rml_relation_db_t *rml_relation_db_end;

/* current execution range */
rml_current_execution_loc_t rml_current_execution_loc;

#define RML_RANGE_EQ  0
#define RML_RANGE_GT  1
#define RML_RANGE_LT -1

/* returns: (we compare only the first line/column of each range
  0 if equal
 +1 if [overlapping] bigger
 -1 if [overlapping] smaller
*/

int rml_range_cmp(rml_range_db_t *r1, rml_range_db_t *r2)
{
	int factor = 1000;
	int t1s,t1e,t2s,t2e;
	t1s = r1->sl*factor + r1->sc; t1e = r1->el*factor + r1->ec; 
	t2s = r2->sl*factor + r2->sc; t2e = r2->el*factor + r2->ec;
	if (t1s == t2s && t1e == t2e) return RML_RANGE_EQ;
	if (t1s > t2s) return RML_RANGE_GT;
	if (t1s < t2s) return RML_RANGE_LT;
}

#define RML_RANGE_1IN2 2
#define RML_RANGE_2IN1 -2
/* returns
  0 if equal
 +1 if [overlapping] bigger
 -1 if [overlapping] smaller
 +2 if r1 included in r2
 -2 if r2 included in r1
*/

int rml_range_incl(rml_range_db_t *r1, rml_range_db_t *r2)
{
	int factor = 1000;
	int t1s,t1e,t2s,t2e;
	t1s = r1->sl*factor + r1->sc; t1e = r1->el*factor + r1->ec; 
	t2s = r2->sl*factor + r2->sc; t2e = r2->el*factor + r2->ec;
	if (t1s >= t2s && t1e <= t2e) return RML_RANGE_1IN2;
	if (t1s <= t2s && t1e >= t2e) return RML_RANGE_2IN1;
	return rml_range_cmp(r1,r2);
}

void rml_str_list_print(rml_str_list_t* list);

/*
this function inits the debugger
*/
int rml_debug_init(void)
{ 
	char process_id[20];
	sprintf(process_id, "%d", getpid());
	/* debug the damn parser/lexer */
	/* aadebug = 1; */
	/* aarmldbdebug = 1; */
	printf("\nrmldb@> - RML debugger\n");
	printf("rmldb@> - 2002, 2003, 2004, LIU/IDA/PELAB, adrpo@ida.liu.se\n");  
	printf("rmldb@> - debugging process %s \n", process_id);
	signal (SIGINT, use_quit);
	/* Readline support.  Set both application name and input file. */ 
	rl_readline_name = RMLDB_PROMPT;
	rl_instream = stdin;
	if (!isatty(STDIN_FILENO)) /* returns 1 on success */
	{
		fprintf(stderr, "RML debugger: isatty(fileno(stdin)) failed: %s\n",
			strerror(errno));
		/* rml_exit(0); */
	}
	if (!(rmldb_ttyname = ttyname(STDIN_FILENO)))
	{
		fprintf(stderr, "RML debugger init ttyname(0) failed: %s\n",
			strerror(errno));
		/*rml_exit(0);*/
	}
	else
	{
		printf("rmldb@> - on tty:%s \n", rmldb_ttyname);
	}

	using_history ();

	/* initalize stuff */
	rml_debug_var_in_start = NULL;
	rml_debug_var_in_end = NULL;
	rml_debug_var_out_start = NULL;
	rml_debug_var_out_end = NULL;

	rml_var_db_start = NULL;
	rml_var_db_end = NULL;
	rml_type_db_start = NULL;
	rml_type_db_end = NULL;
	rml_con_db_start = NULL;
	rml_con_db_end = NULL;
	rml_relation_db_start = NULL;
	rml_relation_db_end = NULL;

	rml_current_execution_loc.file[0] = '\0';
	rml_current_execution_loc.relation[0] = '\0';
	rml_current_execution_loc.goal[0] = '\0';
	rml_current_execution_loc.range.sl = 0;
	rml_current_execution_loc.range.sc = 0;
	rml_current_execution_loc.range.el = 0;
	rml_current_execution_loc.range.ec = 0;

	/* parse for the first commands  */
	execution_type=RMLDB_STEP;
	rmldb_parse();
}

/*
adrpo 2002-10 creation
- this function ends the debugging session.
- i didn't find a good time to call it.
*/
int rml_debug_end(void)
{
	printf("\nrmldb@> - breaking just before exit\n");
	rmldb_parse();
	printf("\nrmldb@> - handing the control to the runtime for exit\n");
}

void show_help(void)
{
	printf("---------------------------------- HELP: COMMANDS -----------------------------------\n");
	printf("Seting/unsetting breakpoints:                    b|break|breakpoint [on|off] string\n");
	printf("Clear all breakpoints:                           cl|clear\n");
	printf("Show all breakpoints:                            sh|show\n");
	printf("Print all the backtrace:                         bt|backtrace\n");
	printf("Print filtered backtrace:                        fbt|fbacktrace filter\n");
	printf("Set the maximum of backtrace entries:            maxbt|maxbacktrace integer\n");
	printf("Set the depth of variable printing:              [set] de|depth integer\n");
	printf("Set the execution mode:                          set st|step [on|off]\n");
	printf("Perform one step:                                st|step\n");
	printf("Run the program:                                 ru|run\n");
	printf("Print the current settings:                      stg|settings\n");
	printf("Showing help:                                    he|help\n");
	printf("Printing the status of Modelica runtime:         sts|stat|status\n");
	printf("Set the output to go to file also:               ou|output file {WORK IN PROGRESS}\n");
	printf("Print the names of live variables:               li|live|livevars\n");
	printf("On/Off printing names of livevars each step:     [set] li|live|livevars on|off\n");
	printf("Print the live variable:                         pr|print var_name\n");
	printf("Print sizeof the live variable:                  sz|size|sizeof var_name\n");
	printf("Display the live variable each step:             di|display var_name\n");
	printf("Un-display the live variable :                   ud|undisplay var_name\n");
	printf("Show display variables:                          di|display\n");
	printf("Un-display ALL display variables:                ud|undisplay\n");
	printf("Send the live variable to external viewer:       gr|graph var_name {WORK IN PROGRESS}\n");
	printf("VERY FAST debugging, no backtrace, no livevars:  fa|fast\n");
	printf("Exiting the debugger/program:                    qu|quit|ex|exit|by|bye\n");
	printf("Debugging/Undebugging the debugger parser:       dbgp|debugparse on|off {aadebug=1/0}\n");
	printf("-------------------------------------------------------------------------------------\n");
}

int is_quit(char *line)
{
	if (strcmp(line, "quit") == 0 || 
		strcmp(line, "exit") == 0 ||
		strcmp(line, "bye") == 0  ||
		strcmp(line, "hit the road jack") == 0 ||
		strcmp(line, "go away") == 0
		) return 1;
	else return 0;
}

int is_help(char *line)
{
	if (strcmp(line, "?") == 0 || 
		strcmp(line, "help") == 0
		) return 1;
	else return 0;
}

void add_relation_filter(char *rname)
{
	int i;
	for (i=0; i < number_of_filters; i++)
	{
		if (strcmp(relation_filter[i], rname) == 0) return;
	}
	if (number_of_filters + 1 == RML_DEBUG_MAX_BREAKPOINTS)
	fprintf(stderr, "\nThe debugger can handle only %d breakpoints, clear some", 
					RML_DEBUG_MAX_BREAKPOINTS);
	strncpy(relation_filter[number_of_filters], rname, RML_DEBUG_MAX_STRING);
	number_of_filters++;
	printf("Breakpoint on: [%s] added to breakpoints list.\n", rname); 
}

void del_relation_filter(char *rname)
{
	int i;
	for (i=0; i < number_of_filters; i++)
	{
		if (strcmp(relation_filter[i], rname) == 0) break;
	}
	strcpy(relation_filter[i], relation_filter[number_of_filters-1]);
	number_of_filters--;  
	printf("Breakpoint on: [%s] removed from breakpoints list.\n", rname); 
}
 
void push_node(char *rname)
{
	struct call_stack_node *last;
	int i = 0;
	struct call_stack_node *tmp;

	for(tmp = call_stack; tmp; tmp = tmp->next)
	{
		i++;
	}
	/* get rid of the first entry */
	if (i > max_backtrace_entries)
	{
		tmp = call_stack;
		if (call_stack) call_stack = call_stack->next;
		free(tmp);
	}

	last = (struct call_stack_node*)malloc(sizeof(struct call_stack_node));
	if(call_stack)
		last->depth = last_node->depth + 1;
	else
	{
		call_stack = last;
		last->depth = 0;
	}
	strcpy(last->relation_name, rname);
	last->next = NULL;
	last->prev = last_node;
	if(last_node)
		last_node->next = last;
	last_node = last;
}

void pop_node(void)
{
	struct call_stack_node *tmp;
	do
	{
		tmp = last_node;
		if(last_node->prev)
			last_node->prev->next = NULL;
		else
			last_node = call_stack = NULL;
		last_node = tmp->prev;
		free(tmp);
	}
	while(last_node);
}

void print_backtrace(char* filter)
{
	struct call_stack_node *tmp;
	printf("-------------- STACK ------------\n"); 
	if(!call_stack)
		rmldb_sprintf("No stack.\n");
	for(tmp = call_stack; tmp; tmp = tmp->next)
	{
		if (!filter) 
			rmldb_sprintf("#%d %s \n", tmp->depth, tmp->relation_name);
		else if (strstr(tmp->relation_name, filter) != NULL)
			rmldb_sprintf("#%d %s \n", tmp->depth, tmp->relation_name); 
		/* fprintf(stderr, "#%d %s \n", tmp->depth, tmp->relation_name); */
	}
	printf("---------------------------------\n"); 
}

void rml_debug_add(struct rml_debug_var_node **start_node, 
				   struct rml_debug_var_node **end_node, void* var_name, void* var)
{
	struct rml_debug_var_node *last;

	last = (struct rml_debug_var_node*)malloc(sizeof(struct rml_debug_var_node));
	last->var_name = var_name;
	last->var = var;
	if(*start_node)
		last->depth = ((struct rml_debug_var_node*)*end_node)->depth + 1;
	else
	{
		*start_node = last;
		last->depth = 0;
	}
	last->next = NULL;
	last->prev = *end_node;
	if(*end_node)
		((struct rml_debug_var_node*)*end_node)->next = last;
	*end_node = last;
}

extern void rml_debug_add_var(int direction, void* var_name, void* var)
{
	if (direction == RML_DEBUG_VAR_IN)
		rml_debug_add(&rml_debug_var_in_start, &rml_debug_var_in_end, var_name, var);
	if (direction == RML_DEBUG_VAR_OUT)
		rml_debug_add(&rml_debug_var_out_start, &rml_debug_var_out_end, var_name, var);
}

extern void clear_rml_debug_vars(void)
{
	struct rml_debug_var_node *tmp;
	for(tmp = rml_debug_var_in_start; tmp; tmp = tmp->next)	free(tmp);
	rml_debug_var_in_start = NULL;
	rml_debug_var_in_end = NULL;
	for(tmp = rml_debug_var_out_start; tmp; tmp = tmp->next) free(tmp);
	rml_debug_var_out_start = NULL;
	rml_debug_var_out_end = NULL;
}

void print_rml_debug_vars(void)
{
	struct rml_debug_var_node *tmp;
	printf("\nLive variables:");
	printf("Results:");
	if (!rml_debug_var_out_start) printf("[none available]");
	for(tmp = rml_debug_var_out_start; tmp; tmp = tmp->next)
		printf("[%s] ", RML_STRINGDATA(tmp->var_name));
	printf("   -   Parameters:");
	if (!rml_debug_var_in_start) printf("[none available]");
	for(tmp = rml_debug_var_in_start; tmp; tmp = tmp->next)
		printf("[%s] ", RML_STRINGDATA(tmp->var_name));
	printf("\n");
}

void set_print_rml_debug_vars(int flag)
{
	rml_debug_print_livevars_each_step = flag;
	if (rml_debug_print_livevars_each_step == RML_LIVEVARS_OFF)
		printf("Print livevars each step set to: OFF\n"); 
	else
		printf("Print livevars each step set to: ON\n"); 
}

void rml_print_variable(char* var_name)
{
	int printed = 0;
	struct rml_debug_var_node *tmp;
	rml_type_db_t* type_db;
	printf("Results:");
	for(tmp = rml_debug_var_out_start; tmp; tmp = tmp->next)
		if (strstr(RML_STRINGDATA(tmp->var_name), var_name) != 0)
		{
			printed = 1;
			type_db = rmldb_get_type(var_name);
			if (type_db)
			{
				printf("\nVARIABLE %s HAS TYPE: ", var_name);
				rml_print_type(type_db->type, stdout);
				printf("\n");
				printf(" %s=", var_name);
				rml_var_show(tmp->var, type_db->type, 0);
			}
			else
			{
				printf("\nI couldn't find the type of variable %s! printing without types \n", var_name);
				printf(" %s=", var_name);
				rml_var_print(tmp->var);
			}
		}
	if (!printed) printf("[not in current context]");
	printed = 0;
	printf("\nParameters:");
	for(tmp = rml_debug_var_in_start; tmp; tmp = tmp->next)
		if (strstr(RML_STRINGDATA(tmp->var_name), var_name) != 0)
		{
			printed = 1;
			type_db = rmldb_get_type(var_name);
			if (type_db)
			{
				printf("\nVARIABLE %s HAS TYPE: ", var_name);
				rml_print_type(type_db->type, stdout);
				printf("\n");
				printf(" %s=", var_name);
				rml_var_show(tmp->var, type_db->type, 0);
			}
			else
			{
				printf("\nI couldn't find the type of variable %s! printing without types \n", var_name);
				printf(" %s=", var_name);
				rml_var_print(tmp->var);
			}

		}
	if (!printed) printf("[not in current context]"); 
	printf("\n");
}

void rml_print_sizeof_variable(char* var_name)
{
	int printed = 0;
	struct rml_debug_var_node *tmp;
	printf("Results:");
	for(tmp = rml_debug_var_out_start; tmp; tmp = tmp->next)
		if (strcmp(var_name, RML_STRINGDATA(tmp->var_name)) == 0)
		{
			printf("[sizeof(%s)=%lu] ", var_name, rml_var_sizeof(tmp->var));
			printed = 1;
		}
	if (!printed) printf("[not in current context]");
	printed = 0;
	printf("   -   Parameters:");
	for(tmp = rml_debug_var_in_start; tmp; tmp = tmp->next)
		if (strcmp(var_name, RML_STRINGDATA(tmp->var_name)) == 0)
		{
			printf("[sizeof(%s)=%lu] ", var_name, rml_var_sizeof(tmp->var));
			printed = 1;
		}
	if (!printed) printf("[not in current context]"); 
	printf("\n");
}

extern void rml_display_variable(char* var_name)
{
	int i;
	for (i=0; i < number_of_display_vars; i++)
	{
		if (strcmp(rml_display_vars[i], var_name) == 0) return;
	}
	if (number_of_display_vars + 1 == RML_DEBUG_MAX_DISPLAY_VARS)
	fprintf(stderr, "\nThe debugger can handle only %d display variables, clear some", 
					RML_DEBUG_MAX_DISPLAY_VARS);
	strcpy(rml_display_vars[number_of_display_vars], var_name);
	number_of_display_vars++;
	rml_print_variable(var_name);
	printf("Variable: [%s] added to display variabile list.\n", var_name); 
}

extern void rml_undisplay_variable(char* var_name)
{
	int i;
	for (i=0; i < number_of_display_vars; i++)
	{
		if (strcmp(rml_display_vars[i], var_name) == 0) break;
	}
	strcpy(rml_display_vars[i], rml_display_vars[number_of_display_vars-1]);
	number_of_display_vars--;
	printf("Variable: [%s] removed from display variabile list\n", var_name); 
}

void print_displayvars(void)
{
	int i = 0;
	if (number_of_display_vars == 0) return; /* printf("No display variables are set\n"); */
	for (i=0; i < number_of_display_vars; i++)
	{
		printf("display #%d -> %s\n", i, rml_display_vars[i]);
		rml_print_variable(rml_display_vars[i]);
	}
}

void show_displayvars(void)
{
	int i = 0;
	printf("------ LIST OF DISPLAY VARIABLES ------\n");
	if (number_of_display_vars == 0) printf("No display variables are set\n");
	for (i=0; i < number_of_display_vars; i++)
	{
		printf("#%d -> %s\n", i, rml_display_vars[i]);
	}
}

void clear_displayvars(void)
{
	number_of_display_vars=0;
	printf("List of display variables cleared.\n");
}


extern void rml_view_variable(char* var_name)
{
	printf("Sending variable %s to external viewer is not working yet\n", var_name); 
}


void set_rmldb_output(char *outputFile)
{
	fprintf(stdout, "The output will go to: %s file also. [NOT IMPLEMENTED YET]\n", outputFile);
}

int open_new_file ()
{
	struct stat buf;
	int tty_file_descriptor = open(rmldb_ttyname, O_RDWR);
	if (tty_file_descriptor == -1) 
	{
		fprintf(stderr, "reopen of %s failed: %s\n",
			rmldb_ttyname,
			strerror(errno));
	}
	if( !(stdin = fdopen(tty_file_descriptor, "rw")) ) 
	{
		fprintf(stderr, "fdopen stdin failed: %s\n",
			strerror(errno));
	}
	/*
	if (!fflush(stdin))
	{
		fprintf(stderr, "fstat on stdin failed: %s\n",
			strerror(errno));
	}
	*/
	/*if (!ioctl(0, TIOCSCTTY, NULL))
	{
		ioctl(stderr, "fstat on stdin failed: %s\n",
			strerror(errno));
	}
	if (!fstat(0, &buf))
	{
		fprintf(stderr, "fstat on stdin failed: %s\n",
			strerror(errno));
	}
	/*if (!isatty(STDIN_FILENO)) return 0;*/
	/*printf("Mumu here!");
	getchar();*/
	aain = stdin;
	/* Readline support.  Set both application name and input file. */ 
	rl_readline_name = RMLDB_PROMPT;
	rl_instream = stdin;
	return 1;
}

/* What happens if you press CTRL+C  */

void
use_quit (sig)
int sig;
{
	/* 
	printf ("\n(interrupt) use quit to exit.\n");
	signal (SIGINT, use_quit);
	*/
	printf ("\nCTRL+C encountered. Breaking at next site....\n");
	execution_type = RMLDB_STEP;
	/* exit (0); */
}


int rmldb_parse(void)
{
	int parse_status = 0;
	do
	{
		while (aaparse() != 0) 
		{
			printf("rmldb@> - command error, showing help\n" );
			show_help();
		}
	}
	while (last_command == 0);
	return 0;
}


void rml_set_depth(char* depth)
{
	depth_of_variable_print = atoi(depth);
	printf("Depth of variable print set to: %d successfully\n", depth_of_variable_print);
	printf("Set it to 0 (ZERO) if you want full printing\n");
}

void rml_set_maxstr(char* maxstr)
{
	max_string_variable_print = atoi(maxstr);
	printf("Printing of max: %d chars from strings set succesfully\n", max_string_variable_print);
	printf("Set it to 0 (ZERO) if you want full strings printed\n");
}

void show_breakpoints(void)
{
	int i = 0;
	printf ("---------- CURRENT BREAKPOINTS ---------\n");
	if (number_of_filters == 0) printf("No breakpoints are set\n");
	for (i=0; i < number_of_filters; i++)
	{
		printf("#%d -> %s\n", i, relation_filter[i]);
	}
}

void clear_breakpoints(void)
{
	number_of_filters=0;
	printf("Breakpoints list cleared\n");
}

void set_max_backtrace(char *maxbt)
{
	max_backtrace_entries = atoi(maxbt);
	printf("Maximum number of backtrace entries set to: %d successfully\n", max_backtrace_entries);
}

void print_settings(void)
{
	printf("---------------------CURRENT SETTINGS---------------------------\n");
	printf("max backtrace entries:              %d\n", max_backtrace_entries);
	printf("depth of variable print:            %d\n", depth_of_variable_print);
	printf("cut strings when print at:          %d\n", max_string_variable_print);
	if (execution_type == RMLDB_STEP) 
	printf("execution type:                     step\n");
	if (execution_type == RMLDB_RUN)
	printf("execution type:                     run\n");
	if (rml_debug_print_livevars_each_step)
	printf("print names of livevars each step:  true\n");
	else 
	printf("print names of livevars each step:  false\n");
	printf("Variables printed at each step/breakpoint:\n");
	show_displayvars();
	printf("breakpoints:\n");
	show_breakpoints();
	printf("tty:   %s\n", rmldb_ttyname);
	printf("RML runtime status:\n");
	rml_show_status();
	print_rml_debug_vars();
	printf("----------------------------------------------------------------\n");
}

/* ---- components ---- */
/* create a list from two tys */
extern rmlLISTty_t* make_rmlLISTty(rml_type_t* type1, rml_type_t* type2)
{
	rmlLISTty_t* rmlLISTty_node = (rmlLISTty_t*)malloc(sizeof(rmlLISTty_t));
	rmlLISTty_node->list_start = NULL;
	rmlLISTty_node->list_end = NULL;
	if (type1)
		rml_type_list_push_back(
			&(rmlLISTty_node->list_start),
			&(rmlLISTty_node->list_end),
			type1);
	if (type2)
		rml_type_list_push_back(
			&(rmlLISTty_node->list_start),
			&(rmlLISTty_node->list_end),
			type2);
	if (type1 || type2)
		rmlLISTty_node->length = rmlLISTty_node->list_end->depth + 1;
	else
		rmlLISTty_node->length = 0;
	return rmlLISTty_node;
}

/* appends ty in front of the list */
extern rmlLISTty_t* make_cons_rmlLISTty(rml_type_t* type, rmlLISTty_t* list)
{
	if (type && list)
		rml_type_list_push_front(
			&list->list_start,
			&list->list_end,
			type);
	if (list->list_end) 
		list->length = list->list_end->depth + 1;
	else
		list->length = 0;
	return list;
}

/* make a VARty */
extern rmlVARty_t* make_rmlVARty(char* id)
{
	rmlVARty_t* rmlVARty_node = (rmlVARty_t*)malloc(sizeof(rmlVARty_t));
	rmlVARty_node->id = id;
	return rmlVARty_node;
}

/* make a CONSty */
extern rmlCONSty_t* make_rmlCONSty(rmlLISTty_t* list, char* id)
{
	rmlCONSty_t* rmlCONSty_node = (rmlCONSty_t*)malloc(sizeof(rmlCONSty_t));
	rmlCONSty_node->list = list;
	rmlCONSty_node->id = id;
	return rmlCONSty_node;
}

/* make a TUPLEty */
extern rmlTUPLEty_t* make_rmlTUPLEty(rmlLISTty_t* list)
{
	rmlTUPLEty_t* rmlTUPLEty_node = (rmlTUPLEty_t*)malloc(sizeof(rmlTUPLEty_t));
	rmlTUPLEty_node->list = list;
	return rmlTUPLEty_node;	
}

/* make a RELty */
extern rmlRELty_t* make_rmlRELty(rmlLISTty_t* list1, rmlLISTty_t* list2)
{
	rmlRELty_t* rmlRELty_node = (rmlRELty_t*)malloc(sizeof(rmlRELty_t));
	rmlRELty_node->list1 = list1;
	rmlRELty_node->list2 = list2;
	return rmlRELty_node;	
}

/* makes a type from a component, and casts it according to kind */
extern rml_type_t* make_rml_type(tyKind_t kind, void* component)
{
	rml_type_t* rml_type_node = (rml_type_t*)malloc(sizeof(rml_type_t));
	rml_type_node->kind = kind;
	switch (kind)
	{
	case eLISTty:  rml_type_node->component.l = (rmlLISTty_t*)component; break;
	case eVARty:   rml_type_node->component.v = (rmlVARty_t*)component; break;
	case eCONSty:  rml_type_node->component.c = (rmlCONSty_t*)component; break;
	case eTUPLEty: rml_type_node->component.t = (rmlTUPLEty_t*)component; break;
	case eRELty:   rml_type_node->component.r = (rmlRELty_t*)component; break;
	}
	rml_type_node->depth = 0;
    rml_type_node->prev = 0;
    rml_type_node->next = 0;
	return rml_type_node;
}

extern void rml_type_list_push_back(rml_type_t **start_node, rml_type_t **end_node, rml_type_t *node)
{
	rml_type_t *last;
	last = node;
	if(*start_node)
		last->depth = ((rml_type_t*)*end_node)->depth + 1;
	else { *start_node = last; last->depth = 0; }
	last->next = NULL;	last->prev = *end_node;
	if(*end_node) ((rml_type_t*)*end_node)->next = last;
	*end_node = last;
}

extern void rml_type_list_push_front(rml_type_t **start_node, rml_type_t **end_node, rml_type_t *node)
{
	rml_type_t *first, *tmp;
	first = node;
	first->depth = 0;
	first->next = NULL;
	if(*start_node)
	{
		/* depth++ on all */
		for(tmp = *start_node; tmp; tmp = tmp->next)
		{
			tmp->depth++;
		}
		/* link front */
		first->next = *start_node;
		((rml_type_t*)*start_node)->prev = first;
	}
	else
	{
		if (!*end_node) *end_node = first;
	}
	first->prev = NULL;
	*start_node = first;
}


extern rml_var_db_t* make_rml_var_db(
		char* file,
		rml_range_db_t* range,
		rml_range_db_t* clause_range,
		char* relation,
		char* name,
		rml_type_db_t* type_db)
{
	rml_var_db_t* rml_var_db_node = (rml_var_db_t*)malloc(sizeof(rml_var_db_t));
	rml_var_db_node->file = file;
	rml_var_db_node->range = range;
	rml_var_db_node->clause_range = clause_range;
	rml_var_db_node->relation = relation;
	rml_var_db_node->name = name;
	rml_var_db_node->type_db = type_db;
	rml_var_db_node->depth = 0;
	rml_var_db_node->prev = NULL;
	rml_var_db_node->next = NULL;
	return rml_var_db_node;
}

void rml_var_db_add(rml_var_db_t **start_node, 
				    rml_var_db_t **end_node, 
					rml_var_db_t *node)
{
	rml_var_db_t *last;
	last = node;
	if(*start_node)
		last->depth = ((rml_var_db_t*)*end_node)->depth + 1;
	else { *start_node = last; last->depth = 0; }
	last->next = NULL;	last->prev = *end_node;
	if(*end_node) ((rml_var_db_t*)*end_node)->next = last;
	*end_node = last;
}

extern rml_con_db_t* make_rml_con_db(
		char* file,
		rml_range_db_t* range,
		char* name,
		rml_type_db_t* type_db,
		int constructor)
{
	rml_con_db_t* rml_con_db_node = (rml_con_db_t*)malloc(sizeof(rml_con_db_t));
	rml_con_db_node->file = file;
	rml_con_db_node->range = range;
	rml_con_db_node->name = name;
	rml_con_db_node->type_db = type_db;
	rml_con_db_node->constructor = constructor;
	rml_con_db_node->is_transparent = 0;
	rml_con_db_node->depth = 0;
	rml_con_db_node->prev = NULL;
	rml_con_db_node->next = NULL;
	return rml_con_db_node;
}

void rml_con_db_add(struct rml_con_db **start_node, 
				    struct rml_con_db **end_node, 
					struct rml_con_db *node)
{
	struct rml_con_db *last;
	last = node;
	if(*start_node)
		last->depth = ((struct rml_con_db*)*end_node)->depth + 1;
	else { *start_node = last; last->depth = 0; }
	last->next = NULL;	last->prev = *end_node;
	if(*end_node) ((struct rml_con_db*)*end_node)->next = last;
	*end_node = last;
}

extern rml_type_db_t* make_rml_type_db(
		char* file,
		rml_range_db_t* range,
		char* name)
{
	rml_type_db_t* rml_type_db_node = (rml_type_db_t*)malloc(sizeof(rml_type_db_t));
	rml_type_db_node->file = file;
	rml_type_db_node->name = name;
	rml_type_db_node->range = range;
	rml_type_db_node->type = NULL;
	rml_type_db_node->depth = 0;
	rml_type_db_node->prev = NULL;
	rml_type_db_node->next = NULL;
	return rml_type_db_node;
}

extern rml_type_db_t* make_rml_type_db_t(
		rml_type_t *type)
{
	rml_type_db_t* rml_type_db_node = (rml_type_db_t*)malloc(sizeof(rml_type_db_t));
	rml_type_db_node->file = NULL;
	rml_type_db_node->range = NULL;
	rml_type_db_node->name = NULL;
	rml_type_db_node->type = type;
	rml_type_db_node->depth = 0;
	rml_type_db_node->prev = NULL;
	rml_type_db_node->next = NULL;
	return rml_type_db_node;
}


void rml_type_db_add(struct rml_type_db **start_node, 
				    struct rml_type_db **end_node, 
					struct rml_type_db *node)
{
	struct rml_type_db *last;
	last = node;
	if(*start_node)
		last->depth = ((struct rml_type_db*)*end_node)->depth + 1;
	else { *start_node = last; last->depth = 0; }
	last->next = NULL;	last->prev = *end_node;
	if(*end_node) ((struct rml_type_db*)*end_node)->next = last;
	*end_node = last;
}

extern rml_relation_db_t* make_rml_relation_db(
		char* file,
		rml_range_db_t* range,
		char* name,
		rml_type_db_t* type_db)
{
	rml_relation_db_t* rml_relation_db_node = (rml_relation_db_t*)malloc(sizeof(rml_relation_db_t));
	rml_relation_db_node->file = file;
	rml_relation_db_node->range = range;
	rml_relation_db_node->name = name;
	rml_relation_db_node->type_db = type_db;	
	rml_relation_db_node->depth = 0;
	rml_relation_db_node->prev = NULL;
	rml_relation_db_node->next = NULL;
	return rml_relation_db_node;
}

void rml_relation_db_add(struct rml_relation_db **start_node, 
				    struct rml_relation_db **end_node, 
					struct rml_relation_db *node)
{
	struct rml_relation_db *last;
	last = node;
	if(*start_node)
		last->depth = ((struct rml_relation_db*)*end_node)->depth + 1;
	else { *start_node = last; last->depth = 0; }
	last->next = NULL;	last->prev = *end_node;
	if(*end_node) ((struct rml_relation_db*)*end_node)->next = last;
	*end_node = last;
}

extern rml_str_t* make_rml_str(char* name)
{
	rml_str_t* rml_str_node = (rml_str_t*)malloc(sizeof(rml_str_t));
	rml_str_node->name = name;
	rml_str_node->depth = 0;
	rml_str_node->next = NULL;
	rml_str_node->prev = NULL;
	return rml_str_node;
}

extern void rml_str_list_add(
					rml_str_t **start_node, 
				    rml_str_t **end_node, 
					rml_str_t *node)
{
	rml_str_t *last;
	last = node;
	if(*start_node)
		last->depth = ((rml_str_t*)*end_node)->depth + 1;
	else { *start_node = last; last->depth = 0; }
	last->next = NULL;	last->prev = *end_node;
	if(*end_node) ((rml_str_t*)*end_node)->next = last;
	*end_node = last;
}

void rml_str_list_print(rml_str_list_t* list)
{
	rml_str_t* tmp;
	fprintf(stderr, "=TYPE_LIST(");
	if (list) 
		for (tmp = list->list_start; tmp; tmp = tmp->next)
		{
			if (tmp->next) fprintf(stderr, "%s, ", tmp->name);
			else fprintf(stderr, "%s", tmp->name);
		}
	fprintf(stderr, ")\n");
}

void rml_type_list_to_list(rmlLISTty_t* list, rml_str_list_t* lst)
{
	rml_type_t* tmp;
	for (tmp = list->list_start; tmp; tmp = tmp->next)
		rml_type_to_list(tmp, lst);
}

extern void rml_type_to_list(rml_type_t *type, rml_str_list_t* list)
{
	if (!type) { return; } 
	switch (type->kind)
	{
	case eLISTty:  
		{
			rml_type_list_to_list(type->component.l, list);
			break;
		}
	case eVARty:   
		{
			rml_str_list_add(
				&list->list_start,
				&list->list_end,
				make_rml_str(type->component.v->id)); 
			break;
		}
	case eCONSty:  
		{ 
			rml_type_list_to_list(type->component.c->list, list);
			rml_str_list_add(
				&list->list_start,
				&list->list_end,
				make_rml_str(type->component.c->id));
			break;
		}
	case eTUPLEty: 
		{ 
			rml_type_list_to_list(type->component.t->list, list);
			break;
		}
	case eRELty:   
		{ 
			rml_type_list_to_list(type->component.r->list1, list);
			rml_type_list_to_list(type->component.r->list2, list);			
			break;
		}
	}
}

void rml_print_type_list(rmlLISTty_t* list, char* separator, FILE* out)
{
	rml_type_t* tmp;
	if (!list) { fprintf(out, "()"); return; }
	for (tmp = list->list_start; tmp; tmp = tmp->next)
		if (tmp->next) { rml_print_type(tmp, out); fprintf(out, "%s ", separator); }
		else rml_print_type(tmp, out);
}

void rml_print_type(rml_type_t *type, FILE* out)
{
	if (!type) { fprintf(out,"()"); return; } 
	switch (type->kind)
	{
	case eLISTty:  
		{
			fprintf(out, "LISTty(["); 
			rml_print_type_list(type->component.l, ",", out);
			fprintf(out, "])");
			break;
		}
	case eVARty:   
		{
			fprintf(out, "VARty(%s)", type->component.v->id, out); 
			break;
		}
	case eCONSty:  
		{
			fprintf(out, "CONSty(["); 
			rml_print_type_list(type->component.c->list, "," , out); 
			fprintf(out, "], %s)", type->component.c->id, out); 
			break;
		}
	case eTUPLEty: 
		{
			fprintf(out, "TUPLEty(["); 
			rml_print_type_list(type->component.t->list, "*", out);
			fprintf(out, "])");
			break;
		}
	case eRELty:   
		{
			fprintf(out, "RELty(["); 
			rml_print_type_list(type->component.r->list1, ",",  out);
			fprintf(out, "] => [");
			rml_print_type_list(type->component.r->list2, ",", out);
			fprintf(out, "])");			
			break;
		}
	}
}

extern rml_type_db_t* rmldb_get_type(char* var_name)
{
	rml_var_db_t* tmp;
	for(tmp = rml_var_db_start; tmp; tmp = tmp->next)
	{
		/* rml_current_execution_loc must be included in variable range -> type */
		if (strcmp(rml_current_execution_loc.file, tmp->file) == 0) 
		if (strcmp(rml_current_execution_loc.relation, tmp->relation) == 0)
		if (strcmp(var_name, tmp->name) == 0)
		if (rml_range_incl(&rml_current_execution_loc.range, tmp->clause_range) ==
			RML_RANGE_1IN2)
			return tmp->type_db;
	}
	printf("\n");
	return NULL;
}

char* rml_get_type_id(rml_con_db_t* con_db)
{
	if (con_db->type_db->type->kind == eRELty)
		return con_db->type_db->type->component.r->list2->list_start->component.c->id;
	if (con_db->type_db->type->kind == eCONSty)
		return con_db->type_db->type->component.c->id;
	fprintf(stderr, "ERROR, unexpected type kind %d", con_db->type_db->type->kind);
	return NULL;
}

rmlLISTty_t* rml_get_type_components(rml_con_db_t* con_db)
{
	if (con_db->type_db->type->kind == eRELty)
		return con_db->type_db->type->component.r->list1;
	if (con_db->type_db->type->kind == eCONSty)
		return NULL;
	fprintf(stderr, "ERROR, unexpected type kind %d", con_db->type_db->type->kind);
	return NULL;
}

void order_rml_con_db(void)
{
	/*
   char *file; int constructor; rml_range_db_t* range; char *name;
   rml_type_db_t* type_db; int depth;
	*/
	char *buff;
	int seq;
	rml_con_db_t *i,*j;
	void *tmp;
	for(j = rml_con_db_start; j; j = j->next)
	for(i = j; i; i = i->next)
	{
	    if (!strcmp(i->file, j->file))
		if (rml_range_cmp(j->range, i->range) == RML_RANGE_GT)
		{
			/* XXXX fixme! move arround ->prev ->next not the contents */
			tmp = (void*)i->range;
			i->range = j->range;
			j->range = (rml_range_db_t*)tmp;

			buff = (char*)malloc(strlen(i->name)+1);
			strcpy(buff, i->name);
			free(i->name);
			i->name = (char*)malloc(strlen(j->name)+1);
			strcpy(i->name,j->name);
			free(j->name);
			j->name = buff;

			tmp = (void*)i->type_db;
			i->type_db = j->type_db;
			j->type_db = (rml_type_db_t*)tmp;

			tmp = (void*)i->depth;
			i->depth = j->depth;
			j->depth = (int)tmp;
		}
	}
	/* order over datatypes of the same type */
	seq = 0;
	for(i = rml_con_db_start; i; i = i->next)
	{
		if (i->next)
		{
			if (!strcmp(rml_get_type_id(i), rml_get_type_id(i->next)))
			{
				i->constructor = seq; seq++;
				i->next->constructor = seq;
			}
			else
			{
				if (seq == 0) /* means there is ONLY ONE CONSTRUCTOR in this datatype */
				{
					rmlLISTty_t* rmlLISTty_node = rml_get_type_components(i);
					/* means is not a constant constructor OR a type > 2 */
					/* must be exactly 1 slot and 1 constructor in the datatype
					 * datatype varTys ident = CONSTRUCTOR tys */
					if (rmlLISTty_node != NULL && rmlLISTty_node->length == 1)  
					{
						i->is_transparent = 1;
						/*
						fprintf(stderr,"TRANSPARENT on %s->%s\n", i->name, rml_get_type_id(i));
						*/
					}
				}
				seq = 0;
			}
		}
		else
		{
			if (seq == 0) /* means there is ONLY ONE CONSTRUCTOR in this datatype */
			{
				rmlLISTty_t* rmlLISTty_node = rml_get_type_components(i);
				/* means is not a constant constructor OR a type > 2 */
				/* must be exactly 1 slot and 1 constructor in the datatype
				 * datatype varTys ident = CONSTRUCTOR tys */
				if (rmlLISTty_node != NULL && rmlLISTty_node->length == 1)  
				{
					i->is_transparent = 1;
					/*
					fprintf(stderr,"TRANSPARENT on %s->%s\n", i->name, rml_get_type_id(i));
					*/				
				}
			}
		}
	}
	/*
	fprintf(stderr,"\n---- TYPES----\n");
	for(i = rml_con_db_start; i; i = i->next)
	{
		fprintf(stderr,"%s:%d.%d.%d.%d | %s->%s[%d][",
			i->file,
			i->range->sl,i->range->sc,i->range->el,i->range->ec,
			rml_get_type_id(i), 
			i->name,
			i->constructor);
		rml_print_type_list(rml_get_type_components(i), ",", stderr);
		fprintf(stderr,"]\n");
	}
	*/
}

char* rml_get_con_name(rml_type_t *type, int constructor)
{
	rml_con_db_t *i;
	if (type)
	{
		rml_str_list_t* types;
		types = (rml_str_list_t*)malloc(sizeof(rml_str_list_t));
		types->list_start = types->list_end = NULL;
		rml_type_to_list(type, types);
		/* 
		rml_print_type(type, stderr);
		rml_str_list_print(types);
		*/
		if (types && types->list_start && types->list_start->name)
		{
			for(i = rml_con_db_start; i; i = i->next)
			{
				rml_str_list_t* components;
				components = (rml_str_list_t*)malloc(sizeof(rml_str_list_t));
				components->list_start = components->list_end = NULL;
				rml_type_to_list(i->type_db->type, components);
				/* 
				rml_print_type(i->type_db->type, stderr);
				rml_str_list_print(components); 
				*/
				if (components && 
					components->list_start && 
					types->list_start->name &&
					components->list_end->name)
				{
					if (!strcmp(types->list_start->name,
						components->list_end->name))
					{
						if (i->constructor == constructor)
							return i->name;
						if (i->is_transparent)
							return i->name;
					}
				}
				rml_str_list_free(components);
			}
		}
		rml_str_list_free(types);
	}
	return NULL;
}


rml_type_t* rml_get_type_comp_from_con(rmlLISTty_t* list, int index)
{
	rml_type_t* tmp;
	if (list) 
		for (tmp = list->list_start; tmp; tmp = tmp->next)
		{
			/*
			uncomment to debug
			fprintf(stderr, "\nINDEX=[%d]-DEPTH=[/%d]->TYPE:", index, tmp->depth); rml_print_type(tmp, stderr);
			*/
			if (tmp->depth == index) return tmp;
		}
	return NULL;
}

rml_type_t* rml_get_type_component(rml_type_t *type, int constructor, int index)
{
	rml_con_db_t *i;
	if (type)
	{
		rml_str_list_t* types;
		types = (rml_str_list_t*)malloc(sizeof(rml_str_list_t));
		types->list_start = types->list_end = NULL;
		rml_type_to_list(type, types);
		/*
		rml_print_type(type, stderr);
		rml_str_list_print(types);
		*/
		if (types && types->list_start, types->list_end->name)
		{
			for(i = rml_con_db_start; i; i = i->next)
			{
				rml_str_list_t* components;
				components = (rml_str_list_t*)malloc(sizeof(rml_str_list_t));
				components->list_start = components->list_end = NULL;
				rml_type_to_list(i->type_db->type, components);
				/*
				rml_print_type(i->type_db->type, stderr);
				rml_str_list_print(components);
				*/
				if (components && 
					components->list_start,
					types->list_start->name,
					components->list_end->name)
				{
					if (!strcmp(types->list_start->name,
						components->list_end->name))
					{
						if (i->constructor == constructor)
							return rml_get_type_comp_from_con(rml_get_type_components(i), index);
						if (i->is_transparent)
						{
							return rml_get_type_comp_from_con(rml_get_type_components(i), 0);
						}
					}
				}
				rml_str_list_free(components);
			}
		}
		rml_str_list_free(types);
	}
	return NULL;
}


extern int rmldb_load_db(char* programdb_file)
{
	char line[RML_DEBUG_MAX_STRING];  /* declare a char array */
	rml_var_db_t *rml_var_db_node;
	struct rml_type_db *rml_type_db_node;
	struct rml_con_db *rml_con_db_node;
	struct rml_relation_db *rml_relation_db_node;

	fprintf(stderr, "loading program database from: %s\n", programdb_file);

	aarmldbin = fopen(programdb_file, "r"); 
	/* open a text file for reading */

	if(aarmldbin==NULL) 
	{
		fprintf(stderr, "Error: can't open file %s.\n", programdb_file);
		/* fclose(file); DON'T PASS A NULL POINTER TO fclose !! */
		rml_exit(2);
	}
	else 
	{
		aarmldbrestart(aarmldbin);
		if (aarmldbparse())
		  fprintf(stderr, "parsing of program database for file %s failed!\n", programdb_file); 
		fclose(aarmldbin);
		order_rml_con_db();
		return 0;
	}
}

extern int rml_str_list_get_size(rml_str_list_t* list)
{
	if (!list->list_end) return 0;
	return list->list_end->depth;
}

extern char* rml_str_list_get_index(rml_str_list_t* list, int index)
{
	rml_str_t* tmp;
	for (tmp = list->list_start; tmp; tmp = tmp->next)
		if(tmp->depth == index) return tmp->name;
	return NULL;
}

void rml_str_list_free(rml_str_list_t* list)
{
	rml_str_t* tmp, *temp;
	tmp = list->list_start;
	while (tmp)
	{
		temp = tmp;
		tmp = tmp->next;
		free(temp);
	}
	free(list);
}

void rml_print_increment(int depth)
{
	int d = 0;
	while(d < depth) { rmldb_sprintf(" "); d++; }
}

rml_type_t* rml_type_unfold(rml_type_t* type)
{
	return type->component.c->list->list_start;
}

void rml_var_show(void *p, rml_type_t* type, int depth)
{
	/* un-comment if you need debugging */
	/*
	if (type) 
	{ 
		fprintf(stdout,"\nTYPE="); 
		rml_print_type(type, stdout); 
		fprintf(stdout,"\n");
	}
	else
	{
		fprintf(stdout,"\nTYPE=NULL");
		fprintf(stdout,"\n");
	}
	*/
	rml_print_increment(depth);

	if (!p) /* nothing to print */
	{ 
		rmldb_sprintf ("NULL"); 
		fflush(stdout); 
		return; 
	}
	if( RML_ISIMM(p) ) 
	{
		rmldb_sprintf ("%d:int", RML_UNTAGFIXNUM(p));    
	} 
	else 
	{
		rml_uint_t phdr = RML_GETHDR(p);            
		if( phdr == RML_REALHDR ) 
		{
			rmldb_sprintf ("%f:real", rml_prim_get_real(p));
			fflush(stdout);
		} else 
			if( RML_HDRISSTRING(phdr) ) 
			{
				if (max_string_variable_print)
				{
					if (max_string_variable_print >= RML_HDRSTRLEN(phdr))
						rmldb_sprintf ("\"%s\":string", RML_STRINGDATA(p)); 
					else
					{
						char format[100];
						sprintf(format,"\"%c%c%d%c\":string", '%', '.', max_string_variable_print, 's');
						rmldb_sprintf(format, RML_STRINGDATA(p));
					}
				}
				else rmldb_sprintf ("\"%s\":string", RML_STRINGDATA(p));
				fflush(stdout);
					/* use if neccesarry RML_HDRSTRLEN(phdr) */
			} 
			else 
			{
				if( RML_HDRISSTRUCT(phdr) ) 
				{
					rml_str_list_t* components;
					char* con_name;
					rml_uint_t slots = RML_HDRSLOTS(phdr);
					rml_uint_t constr = RML_HDRCTOR(phdr);
					int suffix = RML_TYPE_SUFFIX_NONE; /* 1-lists, 2-option, 3-lvar */
					if (slots == 0)
					{
						rmldb_sprintf ("NIL");
						fflush(stdout);
						return;
					}
					else
					{
						if (type)
						{
							components = (rml_str_list_t*)malloc(sizeof(rml_str_list_t));
							components->list_start = components->list_end = NULL;
							rml_type_to_list(type, components);
							/*
							rml_print_type(type, stderr);
							rml_str_list_print(components);
							*/
							/*
							RML_STRUCTHDR(slots,ctor);
							#define RML_NILHDR		RML_STRUCTHDR(0,0)
							#define RML_CONSHDR		RML_STRUCTHDR(2,1)
							*/
							if (components->list_end->name)
							{
								if (!strcmp(components->list_end->name, "list"))   
								{
									suffix = RML_TYPE_SUFFIX_LIST;
								}
								if (!strcmp(components->list_end->name, "option")) 
								{
									suffix = RML_TYPE_SUFFIX_OPTION;
								}
								if (!strcmp(components->list_end->name, "lvar"))   
								{
									suffix = RML_TYPE_SUFFIX_LVAR;
								}
							}
						}
						con_name = rml_get_con_name(type, constr);
						if (con_name) rmldb_sprintf ("%s[%d](", con_name, slots);
						else 
						{
							/* if we haven't found the CONSTRUCTOR NAME print it as it is */
							if (slots == 2 && constr == 1) 
							{ 
								/* rmldb_sprintf("RML.cons("); */
								suffix = RML_TYPE_SUFFIX_LIST;
							}
							else 
								if (slots == 1 && constr == 1) 
								{ 
									/* rmldb_sprintf("RML.SOME("); */
									suffix = RML_TYPE_SUFFIX_OPTION;
								}
							else /* fall to the most miserable case */ 
								 rmldb_sprintf ("S(%d)[%d](", constr, slots);
						}
					}
					void **pp = RML_STRUCTDATA(p);
					fflush(stdout);
					/* function definition  - check this here by the variable type
					   if is a function then is RELty as type !!!!! */
					if ((constr == 64 || constr==13) &&
						slots > 1000000) return;
					if( slots != 0 )
					{
						int i = 0; 
						/* treat here lists separately */
						if (suffix == RML_TYPE_SUFFIX_LIST)
						{
							/*
							#define RML_NILHDR		RML_STRUCTHDR(0,0)
							#define RML_CONSHDR		RML_STRUCTHDR(2,1)
							*/
							rmldb_sprintf ("[");
							while (phdr == RML_CONSHDR)
							{
								void **pcar = RML_STRUCTDATA(p);
								void **pcdr = RML_STRUCTDATA(p);
								/* list = RML.cons(car, cdr) | RML.nil */
								/* cdr */
								pcdr++;
								phdr = RML_GETHDR(pcdr);
								rmldb_sprintf ("\n");
								/* print car */
								rml_var_show(*pcar, 
									rml_type_unfold(type), 
									depth+2);
								if (phdr != RML_NILHDR) 
								{
									rmldb_sprintf (",\n");
									/* print cdr */
									rml_var_show(*pcdr, 
										type, 
										depth+2);
								}
							}
							rmldb_sprintf ("]");
							fflush(stdout);
							return;
						}
						/* treat option separate also */
						if (suffix == RML_TYPE_SUFFIX_OPTION)
						{
							if (phdr == RML_NILHDR) /* NONE */
							{
								rmldb_sprintf ("NONE");
							}
							else
							{
								/* phdr = RML_STRUCTHDR(1,1) = SOME(x) */
								void **pcar = RML_STRUCTDATA(p);
								rmldb_sprintf ("SOME(\n");
								pcar++;
								rml_var_show(*pcar, 
									rml_type_unfold(type), 
									depth+5);
								rmldb_sprintf (")");
							}
							fflush(stdout);
							return;
						}
						/* treat option separate also */
						if (suffix == RML_TYPE_SUFFIX_LVAR)
						{
							/* actually should not happen here */
							rmldb_sprintf ("LVAR(????HERE????)");
							return;
						}
						for (i = 0; i < slots; i++)
						{
							/* tree like display please */
							rmldb_sprintf ("\n");
							rml_var_show(*pp++, 
								rml_get_type_component(type, constr, i), 
								depth+1);
							if (i < slots-1) rmldb_sprintf (",");
							fflush(stdout);
						}
						rmldb_sprintf (")"); fflush(stdout);
					}					    
				} 
				else 
				{
					rmldb_sprintf ("LVAR(UNKNOWN)"); fflush(stdout);
				}
			}
	}
}
